package cds;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * <p>Class to drive the tests on Bounded Buffers.</p>
 * 
 */

public class CDS_02 {
    static final int MAXSLEEP_PROD = 50; //Max sleep for producers
    static final int MAXSLEEP_CONS = 50; //Max sleep for consumers

    static final int NUMPRODUCERS = 3; //Number of producers
    static final int NUMCONSUMERS = 2; //Number of consumers
    
    static final int NUMITEMSPERPROD = 6; //Number of items produced by each producer
    static final int NUMITEMSPERCONS = 9; //Number of items consumed by each consumer
    
    static final int CAPACITY = 3; // Buffer capacity
    
    //static final BoundedBufferInterface<String> sharedBuffer = new SemBoundedBuffer<>(CAPACITY);
    static final BoundedBufferInterface<String> sharedBuffer = new CondBoundedBuffer<>(CAPACITY);
    //static final BoundedBufferInterface<String> sharedBuffer = new BQueueBoundedBuffer<>(CAPACITY, true); //last param: fairness
   

    /**
     * @param args the command line arguments; not used
     */
    public static void main(String[] args) {
        
        List<Producer> producers = new ArrayList<>();
        List<Consumer> consumers = new ArrayList<>();

        //all the producer/consumer threads are created
        for (int i = 0; i++ < NUMPRODUCERS; producers.add(new Producer(sharedBuffer)));
        for (int i = 0; i++ < NUMCONSUMERS; consumers.add(new Consumer(sharedBuffer)));
        //all the producer/consumer threads are started
        for(Producer p : producers) p.start();
        for(Consumer c : consumers) c.start();
        //using lambdas:
        //producers.forEach((p)->p.start());
        //consumers.forEach((c)->c.start());
        
        
        //the main thread waits for all the spawn threads to complete:
        try {
            for(Producer p : producers) p.join();
            for(Consumer c : consumers) c.join();
            
            System.out.println("\nYou will visualize this message only if no thread got blocked.");
        }
        catch (InterruptedException e){}

    }  
}



class Producer extends Thread {

    private static final SeqID prodIDmanager=new SeqID();
    private final BoundedBufferInterface myBuffer;
    private int id;
    private static final ThreadLocal<String> prefix=new ThreadLocal<>();

    public Producer(BoundedBufferInterface buffer) {
        this.myBuffer = buffer;
    }

    @Override
    public void run() {
        //id = ThreadSeqID.get(); //This cannot be placed in the constructor!!!
        id=prodIDmanager.get();
        
        prefix.set("");
        for(int j=0;j<id;j++) {
            prefix.set(prefix.get()+"          ");
        }
            
        String message;
        for (int i = 0; i < CDS_02.NUMITEMSPERPROD; i++) {
            message="";
            message=message+"T" + Integer.toString(id) + " put " + Integer.toString(i);
            try {
                Thread.sleep(new java.util.Random().nextInt(CDS_02.MAXSLEEP_PROD));
                
                myBuffer.put(message);
                System.out.println(prefix.get()+message);
            } catch (InterruptedException e) {}
        }
    }
}


class Consumer extends Thread {

    private static final SeqID consIDmanager = new SeqID(0);
    private final BoundedBufferInterface<String> myBuffer;
    private int id;
    private static final ThreadLocal<String> prefix=new ThreadLocal<>();

    public Consumer(BoundedBufferInterface<String> buffer) {
        this.myBuffer = buffer;
    }

    @Override
    public void run() {
        id = consIDmanager.get(); //This cannot be placed in the constructor!!!
        
        prefix.set(". . ");
        for(int j=0;j<CDS_02.NUMPRODUCERS;j++) {
            prefix.set(prefix.get()+". . . . . ");
        }
        for(int j=0;j<id;j++) {
            prefix.set(prefix.get()+". . . . . ");
        }
        prefix.set(prefix.get()+"Cons " + Integer.toString(id) + " got: \"");
            
        String myReading;
        for (int i = 0; i < CDS_02.NUMITEMSPERCONS; i++) {
            try {
                Thread.sleep(new java.util.Random().nextInt(CDS_02.MAXSLEEP_CONS));

                myReading=myBuffer.take();
                System.out.println(prefix.get()+myReading+"\"");
            } catch (InterruptedException e) {System.out.println("INTERRUPTED!");}
        }
    }
}



class SeqID {
  /**
   * The next ID to be assigned
   **/
  private final AtomicInteger nextID;// = new AtomicInteger(0);
  /**
   * My thread-local ID.
   **/
  private final ThreadLocal<Integer> threadID = new ThreadLocal<Integer>(){
    @Override
    protected Integer initialValue() {
      return nextID.getAndIncrement();
    }
  };

  public SeqID(int initial) {
       nextID= new AtomicInteger(initial);    
  }
  
  public SeqID() {
       this(0);   
  }
  
  public int get() {
    return threadID.get();
  }
  public void reset() {
      nextID.set(0);
  }
  
}
