package cds;

import java.util.concurrent.*;

/**
 * <p>Bounded buffer implementation using \Semaphore</p>
 *
 * <p>This implementation has been obtained reworking an original class 
 * by Brian Goetz and Tim Peierls</p>
 *
 * @param <T> Type of the buffered items
 */

public class SemBoundedBuffer<T> implements BoundedBufferInterface<T> {

    private final Semaphore availableItems, availableSpaces;

    private final T[] buffer;
    private int putIndex = 0, takeIndex = 0;

    public SemBoundedBuffer(int capacity) {
        if (capacity <= 0) {
            throw new IllegalArgumentException();
        }
        availableItems = new Semaphore(0);
        availableSpaces = new Semaphore(capacity);
        buffer = (T[]) new Object[capacity]; // just a trick
    }


    @Override
    public void put(T x) throws InterruptedException {
        
        availableSpaces.acquire();
        synchronized(this){
            int i = putIndex;
            buffer[i] = x;
            putIndex = (++i == buffer.length) ? 0 : i;
        }
        availableItems.release();
    }

    @Override
    public T take() throws InterruptedException {
        T item;

        availableItems.acquire();
        synchronized(this){
            int i = takeIndex;
        
            item = buffer[i];
            buffer[i] = null; //?? 
            takeIndex = (++i == buffer.length) ? 0 : i;
        }
        availableSpaces.release();
        
        return item;
    }

}
